# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from spnego._compat import (
    add_metaclass,

    Optional,
    Union,

    IntEnum,
    IntFlag,
)

from spnego._text import (
    text_type,
)

try:
    from gssapi.raw import GSSError
except ImportError as e:
    GSSError = ()

try:
    WinError = WindowsError
except NameError:
    WinError = ()


class NegotiateOptions(IntFlag):
    """Flags that the caller can use to control the negotiation behaviour.

    A list of features as bit flags that the caller can specify when creating the security context. These flags can
    be used on both Windows or Linux but are a no-op on Windows as it should always have the same features available.
    On Linux the features it can implement depend on a wide range of factors like the system libraries/headers that
    are installed, what GSSAPI implementation is present, and what Python libraries are available.

    This is a pretty advanced feature and is mostly a way to control the kerberos to ntlm fallback behaviour on Linux.

    These are the currently implemented feature flags:

    use_sspi:
        Ensures the context proxy used is :class:`spnego.sspi.SSPIProxy`.

    use_gssapi:
        Ensures the context proxy used is :class:`spnego.gss.GSSAPIProxy`.

    use_negotiate:
        Ensures the context proxy used is :class:`spnego.negotiate.NegotiateProxy`.

    use_ntlm:
        Ensures the context proxy used is :class:`spnego.ntlm.NTLMProxy`.

    negotiate_kerberos:
        Will make sure that Kerberos is at least available to try for authentication when using the `negotiate`
        protocol. If Kerberos cannot be used due to the Python gssapi library not being installed then it will raise a
        :class:`spnego.exceptions.FeatureMissingError`. If Kerberos was available but it cannot get a credential or
        create a context then it will just fallback to NTLM auth. If you wish to only use Kerberos with no NTLM
        fallback, set `protocol='kerberos'` when creating the security context.

    session_key:
        Ensure that the authenticated context will be able to return the session key that was negotiated between the
        client and the server. Older versions of `gss-ntlmssp`_ do not expose the functions required to retrieve this
        info so when this feature flag is set then the NTLM fallback process will use a builtin NTLM process and not
        `gss-ntlmssp`_ if the latter is too old to retrieve the session key.

    wrapping_iov:
        The GSSAPI IOV methods are extensions to the Kerberos spec and not implemented or exposed on all platforms,
        macOS is a popular example. If the caller requires the wrap_iov and unwrap_iov methods this will ensure it
        fails fast before the auth has been set up. Unfortunately there is no fallback for this as if the headers
        aren't present for GSSAPI then we can't do anything to fix that. This won't fail if `negotiate` was used and
        NTLM was the chosen protocol as that happens post negotiation.

    wrapping_winrm:
        To created a wrapped WinRM message the IOV extensions are required when using Kerberos auth. Setting this flag
        will skip Kerberos when `protocol='negotiate'` if the IOV headers aren't present and just fallback to NTLM.

    .. _gss-ntlmssp:
        https://github.com/gssapi/gss-ntlmssp
    """
    none = 0x00000000

    # Force a specific provider
    use_sspi = 0x00000001
    use_gssapi = 0x00000002
    use_negotiate = 0x00000004
    use_ntlm = 0x00000008

    negotiate_kerberos = 0x00000010
    session_key = 0x00000020
    wrapping_iov = 0x00000040
    wrapping_winrm = 0x00000080


class FeatureMissingError(Exception):

    @property
    def feature_id(self):
        return self.args[0]

    @property
    def message(self):
        msg = {
            NegotiateOptions.negotiate_kerberos: 'The Python gssapi library is not installed so Kerberos cannot be '
                                                 'negotiated.',

            NegotiateOptions.wrapping_iov: 'The system is missing the GSSAPI IOV extension headers or NTLM is being '
                                           'requested, cannot utilitze wrap_iov and unwrap_iov',

            NegotiateOptions.wrapping_winrm: 'The system is missing the GSSAPI IOV extension headers required for '
                                             'WinRM encryption with Kerberos.',

        }.get(self.feature_id, 'Unknown option flag: %d' % self.feature_id)

        return msg

    def __str__(self):
        return self.message


class ErrorCode(IntEnum):
    """Common error codes for SPNEGO operations.

    Mostly a copy of the `GSS major error codes`_ with the names made more pythonic. Not all codes have a corresponding
    SpnegoError class as they are reserved for the codes that apply to both GSSAPI and SSPI.

    .. _GSS major error codes:
        https://docs.oracle.com/cd/E19683-01/816-1331/reference-4/index.html
    """
    bad_mech = 1  # BadMechanismError
    bad_name = 2  # BadNameError
    # bad_nametype = 3  # No equivalent in SSPI, shouldn't happen in our code as well.
    bad_bindings = 4  # BadBindings
    # bad_status = 5  # Only used in gss_display_status which we don't care about.
    bad_mic = 6  # BadMICError
    no_cred = 7  # NoCredentialError
    no_context = 8  # NoContextError
    invalid_token = 9  # InvalidTokenError
    # invalid_credential = 10  # No real equivalent in SSPI, shouldn't happen in our code.
    credentials_expired = 11  # CredentialsExpiredError
    context_expired = 12  # ContextExpiredError
    failure = 13  # This is a generic error with the error coming from the minor code, uses SpnegoError directly.
    bad_qop = 14  # UnsupportedQop
    # unauthorized = 15  # Shouldn't happen in our code, we don't do any authorization functions.
    unavailable = 16  # OperationNotAvailableError
    # duplicate_element = 17  # Shouldn't happen in our code.
    # name_not_mn = 18  # Shouldn't happen in our code.


# Implementation is inspired by the python-gssapi project https://github.com/pythongssapi/python-gssapi.
# https://github.com/pythongssapi/python-gssapi/blob/826c02de1c1885896924bf342c60087f369c6b1a/gssapi/raw/misc.pyx#L180
class _SpnegoErrorRegistry(type):
    __registry = {}
    __gssapi_map = {}
    __sspi_map = {}

    def __init__(cls, name, bases, attributes):
        # Load up the registry with the instantiated class so we can look it up when creating a SpnegoError.
        error_code = getattr(cls, 'ERROR_CODE', None)

        if error_code is not None and error_code not in cls.__registry:
            cls.__registry[error_code] = cls

        # Map the system error codes to the common spnego error code.
        for system_attr, mapping in [('_GSSAPI_CODE', cls.__gssapi_map), ('_SSPI_CODE', cls.__sspi_map)]:
            codes = getattr(cls, system_attr, None)

            if codes is None:
                continue

            if not isinstance(codes, (list, tuple)):
                codes = [codes]

            for c in codes:
                mapping[c] = error_code

    def __call__(cls, error_code=None, base_error=None, *args, **kwargs):
        error_code = error_code if error_code is not None else getattr(cls, 'ERROR_CODE', None)

        if error_code is None:
            if not base_error:
                raise ValueError("%s requires either an error_code or base_error" % cls.__name__)

            # GSSError
            if hasattr(base_error, 'maj_code'):
                error_code = cls.__gssapi_map.get(base_error.maj_code, None)

            # WindowsError
            elif hasattr(base_error, 'winerror'):
                error_code = cls.__sspi_map.get(base_error.winerror, None)

            else:
                raise ValueError("base_error of type '%s' is not supported, must be a gssapi.exceptions.GSSError or "
                                 "WindowsError" % type(base_error).__name__)

        new_cls = cls.__registry.get(error_code, cls)
        return super(_SpnegoErrorRegistry, new_cls).__call__(error_code, base_error, *args, **kwargs)


@add_metaclass(_SpnegoErrorRegistry)
class SpnegoError(Exception):
    """Common error for SPNEGO exception.

    Creates an common error record for SPNEGO errors raised by pyspnego. This error record can wrap system level error
    records raised by GSSAPI or SSPI and wrap them into a common error record across the various platforms. While this
    reflects the GSSAPI major codes that can be raised, it is up to the GSSAPI platform to conform to those error
    codes. Some platforms like MIT krb5 always report `GSS_S_FAILURE` and use the minor code to report the actual
    error message.

    Args:
        error_code: The ErrorCode for the error, this must be set if base_error is not set.
        base_error: The system level error from SSPI or GSSAPI, this must be set if error_code is not set.
        context_msg: Optional message to provide more context around the error.

    Attributes:
        base_error (Optional[Union[GSSError, WinError]]): The system level error if one was provided.
    """

    # Classes the subclass this type need to provide the following class attribute:
    #
    # ERROR_CODE = common ErrorCode value for the exception
    # _BASE_MESSAGE = common string that explains the error code in the absence of the system error message.
    #
    # The following attributes are used to map specific system error codes to the common ErrorCode error.
    # _GSSAPI_CODE = The GSSAPI major_code from GSSError to map to the common error code
    # _SSPI_CODE = The winerror value from an WindowsError to map to the common error code

    def __init__(self, error_code=None, base_error=None, context_msg=None):
        self.base_error = base_error  # type: Optional[Union[GSSError, WinError]]
        self._error_code = error_code  # type: Optional[ErrorCode]
        self._context_message = context_msg  # type: Optional[text_type]

        super(SpnegoError, self).__init__(self.message)

    @property
    def message(self):
        error_code = self._error_code if self._error_code is not None else 0xFFFFFFFF

        if self.base_error:
            base_message = str(self.base_error)

        else:
            base_message = getattr(self, '_BASE_MESSAGE', 'Unknown error code')

        msg = "SpnegoError (%d): %s" % (error_code, base_message)
        if self._context_message:
            msg += ", Context: %s" % self._context_message

        return msg


class BadMechanismError(SpnegoError):
    ERROR_CODE = ErrorCode.bad_mech

    _BASE_MESSAGE = "An unsupported mechanism was requested"
    _GSSAPI_CODE = 65536  # GSS_S_BAD_MECH
    _SSPI_CODE = -2146893051  # SEC_E_SECPKG_NOT_FOUND


class BadNameError(SpnegoError):
    ERROR_CODE = ErrorCode.bad_name

    _BASE_MESSAGE = "An invalid name was supplied"
    _GSSAPI_CODE = 1310722  # GSS_S_BAD_NAME
    _SSPI_CODE = -2146893053  # SEC_E_TARGET_UNKNOWN


class BadBindingsError(SpnegoError):
    ERROR_CODE = ErrorCode.bad_bindings

    _BASE_MESSAGE = "Invalid channel bindings"
    _GSSAPI_CODE = 262144  # GSS_BAD_BINDINGS
    _SSPI_CODE = -2146892986  # SEC_E_BAD_BINDINGS


class BadMICError(SpnegoError):
    ERROR_CODE = ErrorCode.bad_mic

    _BASE_MESSAGE = "A token had an invalid Message Integrity Check (MIC)"
    _GSSAPI_CODE = 3932166  # GSS_BAD_MIC
    _SSPI_CODE = -2146893041  # SEC_E_MESSAGE_ALTERED


class NoCredentialError(SpnegoError):
    ERROR_CODE = ErrorCode.no_cred

    _BASE_MESSAGE = "No credentials were supplied, or the credentials were unavailable or inaccessible"
    _GSSAPI_CODE = 458752  # GSS_NO_CRED
    _SSPI_CODE = -2146893042  # SEC_E_NO_CREDENTIALS


class NoContextError(SpnegoError):
    ERROR_CODE = ErrorCode.no_context

    _BASE_MESSAGE = "No context has been established, or invalid handled passed in"
    _GSSAPI_CODE = 524288  # GSS_S_NO_CONTEXT
    _SSPI_CODE = -2146893055  # SEC_E_INVALID_HANDLE


class InvalidTokenError(SpnegoError):
    ERROR_CODE = ErrorCode.invalid_token

    _BASE_MESSAGE = "A token was invalid, or the logon was denied"
    _GSSAPI_CODE = 589824  # GSS_S_DEFECTIVE_TOKEN
    _SSPI_CODE = [-2146893044, -2146893048]  # SEC_E_LOGON_DENIED, SEC_E_INVALID_TOKEN


class CredentialsExpiredError(SpnegoError):
    ERROR_CODE = ErrorCode.credentials_expired

    _BASE_MESSAGE = "The referenced credentials have expired"
    _GSSAPI_CODE = 720896  # * GSS_S_CREDENTIALS_EXPIRED
    _SSPI_CODE = -1073741711  # STATUS_PASSWORD_EXPIRED


class ContextExpiredError(SpnegoError):
    ERROR_CODE = ErrorCode.context_expired

    _BASE_MESSAGE = "Security context has expired"
    _GSSAPI_CODE = 786432  # GSS_S_CONTEXT_EXPIRED
    _SSPI_CODE = -2146893033  # SEC_E_CONTEXT_EXPIRED


class UnsupportedQop(SpnegoError):
    ERROR_CODE = ErrorCode.bad_qop

    _BASE_MESSAGE = "The quality-of-protection requested could not be provided"
    _GSSAPI_CODE = 917504  # GSS_S_BAD_QOP
    _SSPI_CODE = -2146893046  # SEC_E_QOP_NOT_SUPPORTED


class OperationNotAvailableError(SpnegoError):
    ERROR_CODE = ErrorCode.unavailable

    _BASE_MESSAGE = "Operation not supported or available"
    _GSSAPI_CODE = 1048576  # GSS_S_UNAVAILABLE
    _SSPI_CODE = -2146893054  # SEC_E_UNSUPPORTED_FUNCTION



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from spnego._sspi_raw.windows cimport (
    LONG,
    LPWSTR,
    PVOID,
    WCHAR
)


cdef extern from "Security.h":
    # Types
    ctypedef LONG SECURITY_STATUS
    ctypedef WCHAR SEC_WCHAR;

    # Defs
    cdef unsigned long SECBUFFER_VERSION

    cdef unsigned long SECBUFFER_EMPTY
    cdef unsigned long SECBUFFER_DATA
    cdef unsigned long SECBUFFER_TOKEN
    cdef unsigned long SECBUFFER_PKG_PARAMS
    cdef unsigned long SECBUFFER_MISSING
    cdef unsigned long SECBUFFER_EXTRA
    cdef unsigned long SECBUFFER_STREAM_TRAILER
    cdef unsigned long SECBUFFER_STREAM_HEADER
    cdef unsigned long SECBUFFER_NEGOTIATION_INFO
    cdef unsigned long SECBUFFER_PADDING
    cdef unsigned long SECBUFFER_STREAM
    cdef unsigned long SECBUFFER_MECHLIST
    cdef unsigned long SECBUFFER_MECHLIST_SIGNATURE
    cdef unsigned long SECBUFFER_TARGET
    cdef unsigned long SECBUFFER_CHANNEL_BINDINGS
    cdef unsigned long SECBUFFER_CHANGE_PASS_RESPONSE
    cdef unsigned long SECBUFFER_TARGET_HOST

    cdef unsigned long SECBUFFER_ATTRMASK
    cdef unsigned long SECBUFFER_READONLY
    cdef unsigned long SECBUFFER_READONLY_WITH_CHECKSUM
    cdef unsigned long SECBUFFER_RESERVED

    cdef unsigned long SECURITY_NATIVE_DREP
    cdef unsigned long SECURITY_NETWORK_DREP

    cdef unsigned long SECPKG_ATTR_NAMES
    cdef unsigned long SECPKG_ATTR_PACKAGE_INFO
    cdef unsigned long SECPKG_ATTR_SIZES
    cdef unsigned long SECPKG_ATTR_SESSION_KEY

    cdef unsigned long SECPKG_CRED_INBOUND
    cdef unsigned long SECPKG_CRED_OUTBOUND
    cdef unsigned long SECPKG_CRED_BOTH
    cdef unsigned long SECPKG_CRED_DEFAULT
    cdef unsigned long SECPKG_CRED_RESERVED

    cdef unsigned long ASC_REQ_DELEGATE
    cdef unsigned long ASC_REQ_MUTUAL_AUTH
    cdef unsigned long ASC_REQ_REPLAY_DETECT
    cdef unsigned long ASC_REQ_SEQUENCE_DETECT
    cdef unsigned long ASC_REQ_CONFIDENTIALITY
    cdef unsigned long ASC_REQ_USE_SESSION_KEY
    cdef unsigned long ASC_REQ_ALLOCATE_MEMORY
    cdef unsigned long ASC_REQ_USE_DCE_STYLE
    cdef unsigned long ASC_REQ_DATAGRAM
    cdef unsigned long ASC_REQ_CONNECTION
    cdef unsigned long ASC_REQ_CALL_LEVEL
    cdef unsigned long ASC_REQ_FRAGMENT_SUPPLIED
    cdef unsigned long ASC_REQ_EXTENDED_ERROR
    cdef unsigned long ASC_REQ_STREAM
    cdef unsigned long ASC_REQ_INTEGRITY
    cdef unsigned long ASC_REQ_LICENSING
    cdef unsigned long ASC_REQ_IDENTIFY
    cdef unsigned long ASC_REQ_ALLOW_NULL_SESSION
    cdef unsigned long ASC_REQ_ALLOW_NON_USER_LOGONS
    cdef unsigned long ASC_REQ_ALLOW_CONTEXT_REPLAY
    cdef unsigned long ASC_REQ_FRAGMENT_TO_FIT
    cdef unsigned long ASC_REQ_NO_TOKEN
    cdef unsigned long ASC_REQ_PROXY_BINDINGS
    cdef unsigned long ASC_REQ_ALLOW_MISSING_BINDINGS

    cdef unsigned long ASC_RET_DELEGATE
    cdef unsigned long ASC_RET_MUTUAL_AUTH
    cdef unsigned long ASC_RET_REPLAY_DETECT
    cdef unsigned long ASC_RET_SEQUENCE_DETECT
    cdef unsigned long ASC_RET_CONFIDENTIALITY
    cdef unsigned long ASC_RET_USE_SESSION_KEY
    cdef unsigned long ASC_RET_ALLOCATED_MEMORY
    cdef unsigned long ASC_RET_USED_DCE_STYLE
    cdef unsigned long ASC_RET_DATAGRAM
    cdef unsigned long ASC_RET_CONNECTION
    cdef unsigned long ASC_RET_CALL_LEVEL
    cdef unsigned long ASC_RET_THIRD_LEG_FAILED
    cdef unsigned long ASC_RET_EXTENDED_ERROR
    cdef unsigned long ASC_RET_STREAM
    cdef unsigned long ASC_RET_INTEGRITY
    cdef unsigned long ASC_RET_LICENSING
    cdef unsigned long ASC_RET_IDENTIFY
    cdef unsigned long ASC_RET_NULL_SESSION
    cdef unsigned long ASC_RET_ALLOW_NON_USER_LOGONS
    cdef unsigned long ASC_RET_ALLOW_CONTEXT_REPLAY
    cdef unsigned long ASC_RET_FRAGMENT_ONLY
    cdef unsigned long ASC_RET_NO_TOKEN
    cdef unsigned long ASC_RET_NO_ADDITIONAL_TOKEN

    cdef unsigned long ISC_REQ_DELEGATE
    cdef unsigned long ISC_REQ_MUTUAL_AUTH
    cdef unsigned long ISC_REQ_REPLAY_DETECT
    cdef unsigned long ISC_REQ_SEQUENCE_DETECT
    cdef unsigned long ISC_REQ_CONFIDENTIALITY
    cdef unsigned long ISC_REQ_USE_SESSION_KEY
    cdef unsigned long ISC_REQ_PROMPT_FOR_CREDS
    cdef unsigned long ISC_REQ_USE_SUPPLIED_CREDS
    cdef unsigned long ISC_REQ_ALLOCATE_MEMORY
    cdef unsigned long ISC_REQ_USE_DCE_STYLE
    cdef unsigned long ISC_REQ_DATAGRAM
    cdef unsigned long ISC_REQ_CONNECTION
    cdef unsigned long ISC_REQ_CALL_LEVEL
    cdef unsigned long ISC_REQ_FRAGMENT_SUPPLIED
    cdef unsigned long ISC_REQ_EXTENDED_ERROR
    cdef unsigned long ISC_REQ_STREAM
    cdef unsigned long ISC_REQ_INTEGRITY
    cdef unsigned long ISC_REQ_IDENTIFY
    cdef unsigned long ISC_REQ_NULL_SESSION
    cdef unsigned long ISC_REQ_MANUAL_CRED_VALIDATION
    cdef unsigned long ISC_REQ_RESERVED1
    cdef unsigned long ISC_REQ_FRAGMENT_TO_FIT
    cdef unsigned long ISC_REQ_FORWARD_CREDENTIALS
    cdef unsigned long ISC_REQ_NO_INTEGRITY
    cdef unsigned long ISC_REQ_USE_HTTP_STYLE
    cdef unsigned long ISC_REQ_UNVERIFIED_TARGET_NAME
    cdef unsigned long ISC_REQ_CONFIDENTIALITY_ONLY

    cdef unsigned long ISC_RET_DELEGATE
    cdef unsigned long ISC_RET_MUTUAL_AUTH
    cdef unsigned long ISC_RET_REPLAY_DETECT
    cdef unsigned long ISC_RET_SEQUENCE_DETECT
    cdef unsigned long ISC_RET_CONFIDENTIALITY
    cdef unsigned long ISC_RET_USE_SESSION_KEY
    cdef unsigned long ISC_RET_USED_COLLECTED_CREDS
    cdef unsigned long ISC_RET_USED_SUPPLIED_CREDS
    cdef unsigned long ISC_RET_ALLOCATED_MEMORY
    cdef unsigned long ISC_RET_USED_DCE_STYLE
    cdef unsigned long ISC_RET_DATAGRAM
    cdef unsigned long ISC_RET_CONNECTION
    cdef unsigned long ISC_RET_INTERMEDIATE_RETURN
    cdef unsigned long ISC_RET_CALL_LEVEL
    cdef unsigned long ISC_RET_EXTENDED_ERROR
    cdef unsigned long ISC_RET_STREAM
    cdef unsigned long ISC_RET_INTEGRITY
    cdef unsigned long ISC_RET_IDENTIFY
    cdef unsigned long ISC_RET_NULL_SESSION
    cdef unsigned long ISC_RET_MANUAL_CRED_VALIDATION
    cdef unsigned long ISC_RET_RESERVED1
    cdef unsigned long ISC_RET_FRAGMENT_ONLY
    cdef unsigned long ISC_RET_FORWARD_CREDENTIALS
    cdef unsigned long ISC_RET_USED_HTTP_STYLE
    cdef unsigned long ISC_RET_NO_ADDITIONAL_TOKEN
    cdef unsigned long ISC_RET_REAUTHENTICATION

    cdef unsigned long SEC_WINNT_AUTH_IDENTITY_ANSI
    cdef unsigned long SEC_WINNT_AUTH_IDENTITY_UNICODE

    cdef unsigned long SECQOP_WRAP_NO_ENCRYPT

    cdef SECURITY_STATUS SEC_I_COMPLETE_AND_CONTINUE
    cdef SECURITY_STATUS SEC_I_COMPLETE_NEEDED
    cdef SECURITY_STATUS SEC_I_CONTINUE_NEEDED
    cdef SECURITY_STATUS SEC_I_INCOMPLETE_CREDENTIALS
    cdef SECURITY_STATUS SEC_E_INCOMPLETE_MESSAGE
    cdef SECURITY_STATUS SEC_E_OK
    cdef SECURITY_STATUS SEC_E_INSUFFICIENT_MEMORY
    cdef SECURITY_STATUS SEC_E_INTERNAL_ERROR
    cdef SECURITY_STATUS SEC_E_INVALID_HANDLE
    cdef SECURITY_STATUS SEC_E_INVALID_TOKEN
    cdef SECURITY_STATUS SEC_E_LOGON_DENIED
    cdef SECURITY_STATUS SEC_E_NO_AUTHENTICATING_AUTHORITY
    cdef SECURITY_STATUS SEC_E_NO_CREDENTIALS
    cdef SECURITY_STATUS SEC_E_TARGET_UNKNOWN
    cdef SECURITY_STATUS SEC_E_UNSUPPORTED_FUNCTION
    cdef SECURITY_STATUS SEC_E_WRONG_PRINCIPAL
    cdef SECURITY_STATUS SEC_E_NOT_OWNER
    cdef SECURITY_STATUS SEC_E_SECPKG_NOT_FOUND
    cdef SECURITY_STATUS SEC_E_UNKNOWN_CREDENTIALS
    cdef SECURITY_STATUS SEC_E_BUFFER_TOO_SMALL
    cdef SECURITY_STATUS SEC_E_CRYPTO_SYSTEM_INVALID
    cdef SECURITY_STATUS SEC_E_MESSAGE_ALTERED
    cdef SECURITY_STATUS SEC_E_OUT_OF_SEQUENCE
    cdef SECURITY_STATUS SEC_E_QOP_NOT_SUPPORTED
    cdef SECURITY_STATUS SEC_E_CONTEXT_EXPIRED

    # Structs
    struct _SEC_WINNT_AUTH_IDENTITY_W:
        void *User;
        unsigned long UserLength;
        void *Domain;
        unsigned long DomainLength;
        void *Password;
        unsigned long PasswordLength;
        unsigned long Flags;
    ctypedef _SEC_WINNT_AUTH_IDENTITY_W SEC_WINNT_AUTH_IDENTITY_W
    ctypedef SEC_WINNT_AUTH_IDENTITY_W *PSEC_WINNT_AUTH_IDENTITY_W;

    struct _SecBuffer:
        unsigned long cbBuffer
        unsigned long BufferType
        void *pvBuffer
    ctypedef _SecBuffer SecBuffer
    ctypedef SecBuffer *PSecBuffer

    struct _SecBufferDesc:
        unsigned long ulVersion
        unsigned long cBuffers
        PSecBuffer pBuffers
    ctypedef _SecBufferDesc SecBufferDesc
    ctypedef SecBufferDesc *PSecBufferDesc

    struct _SecHandle:
        pass
    ctypedef _SecHandle SecHandle
    ctypedef SecHandle *PSecHandle

    ctypedef SecHandle CredHandle
    ctypedef PSecHandle PCredHandle

    ctypedef SecHandle CtxtHandle
    ctypedef PSecHandle PCtxtHandle

    struct _SecPkgContext_Names:
        SEC_WCHAR *sUserName;
    ctypedef _SecPkgContext_Names SecPkgContext_Names
    ctypedef SecPkgContext_Names *PSecPkgContext_Names

    struct _SecPkgContext_SessionKey:
        unsigned long SessionKeyLength;
        void *SessionKey;
    ctypedef _SecPkgContext_SessionKey SecPkgContext_SessionKey
    ctypedef SecPkgContext_SessionKey *PSecPkgContext_SessionKey

    struct _SecPkgInfoW:
        unsigned long fCapabilities
        unsigned short wVersion
        unsigned short wRPCID
        unsigned long cbMaxToken
        SEC_WCHAR *Name
        SEC_WCHAR *Comment
    ctypedef _SecPkgInfoW SecPkgInfoW
    ctypedef SecPkgInfoW *PSecPkgInfoW

    struct _SecPkgContext_PackageInfoW:
        PSecPkgInfoW PackageInfo
    ctypedef _SecPkgContext_PackageInfoW SecPkgContext_PackageInfoW
    ctypedef SecPkgContext_PackageInfoW *PSecPkgContext_PackageInfoW

    struct _SecPkgContext_Sizes:
        unsigned long cbMaxToken
        unsigned long cbMaxSignature
        unsigned long cbBlockSize
        unsigned long cbSecurityTrailer
    ctypedef _SecPkgContext_Sizes SecPkgContext_Sizes
    ctypedef SecPkgContext_Sizes *PSecPkgContext_Sizes

    ctypedef struct _SECURITY_INTEGER:
        unsigned long LowPart
        long HighPart
    ctypedef _SECURITY_INTEGER SECURITY_INTEGER
    ctypedef SECURITY_INTEGER *PSECURITY_INTEGER
    ctypedef SECURITY_INTEGER TimeStamp
    ctypedef SECURITY_INTEGER *PTimeStamp

    # Functions
    SECURITY_STATUS __stdcall AcceptSecurityContext(
        PCredHandle    phCredential,
        PCtxtHandle    phContext,
        PSecBufferDesc pInput,
        unsigned long  fContextReq,
        unsigned long  TargetDataRep,
        PCtxtHandle    phNewContext,
        PSecBufferDesc pOutput,
        unsigned long  *pfContextAttr,
        PTimeStamp     ptsExpiry
    )

    SECURITY_STATUS __stdcall AcquireCredentialsHandleW(
        LPWSTR           pPrincipal,
        LPWSTR           pPackage,
        unsigned long    fCredentialUse,
        void             *pvLogonId,
        void             *pAuthData,
        void             *pGetKeyFn,
        void             *pvGetKeyArgument,
        PCredHandle      phCredential,
        PTimeStamp       ptsExpiry
    )

    SECURITY_STATUS __stdcall CompleteAuthToken(
        PCtxtHandle    phContext,
        PSecBufferDesc pToken
    )

    SECURITY_STATUS __stdcall DecryptMessage(
        PCtxtHandle    phContext,
        PSecBufferDesc pMessage,
        unsigned long  MessageSeqNo,
        unsigned long  *pfQOP
    )

    SECURITY_STATUS __stdcall DeleteSecurityContext(
        PCtxtHandle phContext
    )

    SECURITY_STATUS __stdcall EncryptMessage(
        PCtxtHandle    phContext,
        unsigned long  fQOP,
        PSecBufferDesc pMessage,
        unsigned long  MessageSeqNo
    )

    SECURITY_STATUS __stdcall FreeContextBuffer(
        PVOID pvContextBuffer
    )

    SECURITY_STATUS __stdcall FreeCredentialsHandle(
        PCredHandle phCredential
    )

    SECURITY_STATUS __stdcall InitializeSecurityContextW(
        PCredHandle      phCredential,
        PCtxtHandle      phContext,
        LPWSTR           pTargetName,
        unsigned long    fContextReq,
        unsigned long    Reserved1,
        unsigned long    TargetDataRep,
        PSecBufferDesc   pInput,
        unsigned long    Reserved2,
        PCtxtHandle      phNewContext,
        PSecBufferDesc   pOutput,
        unsigned long    *pfContextAttr,
        PTimeStamp       ptsExpiry
    )

    SECURITY_STATUS __stdcall MakeSignature(
        PCtxtHandle    phContext,
        unsigned long  fQOP,
        PSecBufferDesc pMessage,
        unsigned long  MessageSeqNo
    )

    SECURITY_STATUS __stdcall QueryContextAttributesW(
        PCtxtHandle   phContext,
        unsigned long ulAttribute,
        void          *pBuffer
    )

    SECURITY_STATUS __stdcall VerifySignature(
        PCtxtHandle    phContext,
        PSecBufferDesc pMessage,
        unsigned long  MessageSeqNo,
        unsigned long  *pfQOP
    )



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type  # noqa (fixes E402 for the imports below)

import collections

from spnego._compat import (
    IntFlag,
)


class BufferType(IntFlag):
    """Buffer types to use for an IOVBuffer type.

    These are the IOVBuffer type flags that can be set for an IOVBuffer. The keys are a generified name for the
    respective SSPI and GSSAPI flags.
    """
    empty = 0  # SECBUFFER_EMPTY | GSS_IOV_BUFFER_TYPE_EMPTY
    data = 1  # SECBUFFER_DATA | GSS_IOV_BUFFER_TYPE_DATA
    header = 2  # SECBUFFER_TOKEN | GSS_IOV_BUFFER_TYPE_HEADER
    pkg_params = 3  # SECBUFFER_PKG_PARAMS | GSS_IOV_BUFFER_TYPE_MECH_PARAMS
    trailer = 7  # SECBUFFER_STREAM_HEADER | GSS_IOV_BUFFER_TYPE_TRAILER
    padding = 9  # SECBUFFER_PADDING | GSS_IOV_BUFFER_TYPE_PADDING
    stream = 10  # SECBUFFER_STREAM | GSS_IOV_BUFFER_TYPE_STREAM
    sign_only = 11  # SECBUFFER_MECHLIST | GSS_IOV_BUFFER_TYPE_SIGN_ONLY
    mic_token = 12  # SECBUFFER_MECHLIST_SIGNATURE | GSS_IOV_BUFFER_TYPE_MIC_TOKEN


IOVBuffer = collections.namedtuple('IOVBuffer', ['type', 'data'])
"""A buffer to pass as a list to :meth:`wrap_iov()`.

Defines the buffer inside a list that is passed to :meth:`wrap_iov()`. A list of these buffers are also returned in the
`IOVUnwrapResult` under the `buffers` attribute.

On SSPI only a buffer of the type `header`, `trailer`, or `padding` can be auto allocated. On GSSAPI all buffers can be
auto allocated when `data=True` but the behaviour behind this is dependent on the mech it is run for.

Attributes:
    type (BufferType): The type of the IOV buffer.
    data (Union[bytes, int, bool]): On the output from the `*_iov` functions this is the bytes buffer or `None` if the
        buffer wasn't set. When used as an input to the `*_iov` functions this can be the buffer bytes, the length of
        buffer to allocate or a bool to state whether the buffer is auto allocated or not.
"""



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type  # noqa (fixes E402 for the imports below)

__version__ = '0.1.4'



# Copyright: (c) 2020, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdlib cimport (
    free,
    malloc,
)

from spnego._sspi_raw.windows cimport (
    CP_UTF8,
    LPWSTR,
    MultiByteToWideChar,
    WCHAR,
    wchar_t
)


cdef extern from "ctype.h":
  int wcslen(wchar_t*)


cdef extern from "Python.h":
    PyUnicode_FromWideChar(const wchar_t *w, Py_ssize_t size)


cdef class WideChar:

    def __cinit__(WideChar self, size_t length):
        self.length = length
        self.buffer = NULL

        if length:
            self.buffer = <LPWSTR>malloc(length * sizeof(WCHAR))
            if not self.buffer:
                raise MemoryError("Cannot malloc for WideChar")

    def __len__(WideChar self):
        return self.length

    def __dealloc__(WideChar self):
        if self.buffer != NULL:
            free(self.buffer)

    def to_text(WideChar self, size_t length=0):
        if not length and self.length == 0:
            return u""

        # Subtract from self.length to remove the null char that LPWSTR points to.
        return u16_to_text(self.buffer, length if length else self.length - 1)

    @staticmethod
    def from_text(unicode text):
        if not text:
            return WideChar(0)

        b_text = text.encode('utf-8', 'strict')

        # Get the expected length of the text as a wide_char array and allocate it
        length = MultiByteToWideChar(CP_UTF8, 0, b_text, -1, NULL, 0)

        # Create the new WideChar object and set the text to the newly allocated buffer.
        wide_char = WideChar(length)
        MultiByteToWideChar(CP_UTF8, 0, b_text, -1, wide_char.buffer, length)
        return wide_char


cdef unicode u16_to_text(LPWSTR s, size_t length):
    # TODO: pass length straight in when Python 2.7 is dropped.
    if length == -1:
        length = wcslen(s)

    return PyUnicode_FromWideChar(s, length)
